/*
 *	Armadillo-2x0 LED Driver
 *	Copyright (C) 2005-2006 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <asm/io.h>

#include <asm/arch/platform.h>
#include <asm/arch/armadillo2x0_led.h>

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-2x0 LED driver"
#define LICENSE     "GPL v2"

#define LED_MINOR   (215)
#define LED_DEVICE  "led"

#define VERSION     LED_DEVICE ": " DESCRIPTION ", (C) 2005-2006 " AUTHOR

#define ON       (1)
#define OFF      (2)
#define STATUS   (4)

static spinlock_t led_lock = SPIN_LOCK_UNLOCKED;
static unsigned long led_rivalstatus = 0;

#define LED_IS_OPEN    (0x01)
#define LED_READ_COMP  (0x02)
#define LED_WRITE_COMP (0x04)

static void armadillo2x0_led_red_timerfunc (unsigned long data);
static void armadillo2x0_led_green_timerfunc (unsigned long data);
static struct timer_list armadillo2x0_led_red_timer =
TIMER_INITIALIZER( armadillo2x0_led_red_timerfunc, 0, 0);
static struct timer_list armadillo2x0_led_green_timer =
TIMER_INITIALIZER( armadillo2x0_led_green_timerfunc, 0, 0);

#undef DEBUG
#if defined(DEBUG)
#define KDEBUG(args...) printk("led: " args)
#else
#define KDEBUG(args...)
#endif
struct led_info {
	unsigned long ddr;
	unsigned long dr;
	unsigned long bit;
	unsigned int  status;
	unsigned int  blinkstatus;
	struct timer_list *timer;
};
static struct led_info led_red        =
  {GPIO_PCDDR, GPIO_PCDR, 0x00000080, OFF, OFF, &armadillo2x0_led_red_timer};
static struct led_info led_green      =
  {GPIO_PEDDR, GPIO_PEDR, 0x00000001, OFF, OFF, &armadillo2x0_led_green_timer};
static struct led_info led_red_a210   =
  {GPIO_PEDDR, GPIO_PEDR, 0x00000001, OFF, OFF, &armadillo2x0_led_red_timer};
static struct led_info led_green_a210 =
  {GPIO_PEDDR, GPIO_PEDR, 0x00000002, OFF, OFF, &armadillo2x0_led_green_timer};

static int red_enable   = 0;
static int green_enable = 0;
#define DEVCFG_CheckBit \
	(SYSCON_DEVCFG_EonIDE | SYSCON_DEVCFG_GonIDE | SYSCON_DEVCFG_HonIDE)

static void check_platform(void){
	switch(get_board_id()){
	case BOARD_ID_A210:
		red_enable   = 1;
		green_enable = 1;
		led_red      = led_red_a210;
		led_green    = led_green_a210;
		break;
	case BOARD_ID_A210C:
	case BOARD_ID_A220:
	case BOARD_ID_A230:
	case BOARD_ID_A240:
		red_enable   = 1;
		green_enable = 1;
		break;
	default:
		break;
	}
}

static int led_cont(struct led_info *info, int mode)
{
	unsigned long tmp;
	KDEBUG("led_cont\n");

	tmp = inl(info->ddr);
	outl((tmp | info->bit), info->ddr);

	switch(mode){
	case ON:
		tmp = (inl(info->dr) | info->bit);
		outl(tmp, info->dr);
		return 1;

	case OFF:
		tmp = (inl(info->dr) & ~info->bit);
		outl(tmp, info->dr);
		return 0;

	case STATUS:
        	tmp = (inl(info->dr) & info->bit);
		return tmp ? 1 : 0;
	}
	return 0;
}

static int led_on(struct led_info *info)
{
	info->status = ON;
	return led_cont(info, ON);
}

static int led_off(struct led_info *info)
{
	info->status = OFF;
	return led_cont(info, OFF);
}

static inline int led_status(struct led_info *info)
{
	return info->status == ON;
}

static int led_blinkon(struct led_info *info)
{
	if (info->blinkstatus != ON) {
		info->blinkstatus = ON;
		led_cont(info, led_cont(info, STATUS) ? OFF : ON);
		mod_timer(info->timer, jiffies + HZ/2);
	}
	return 1;
}

static int led_blinkoff(struct led_info *info)
{
	if (info->blinkstatus != OFF) {
		info->blinkstatus = OFF;
		del_timer_sync(info->timer);
		led_cont(info, info->status);
	}
	return 0;
}

static inline int led_blinkstatus(struct led_info *info)
{
	return info->blinkstatus == ON;
}

static void armadillo2x0_led_red_timerfunc (unsigned long data)
{
	int tmp;
	tmp = led_cont(&led_red, STATUS) ? OFF : ON;
	led_cont(&led_red, tmp);
	mod_timer(led_red.timer, jiffies + HZ/2);
}

static void armadillo2x0_led_green_timerfunc (unsigned long data)
{
	int tmp;
	tmp = led_cont(&led_green, STATUS) ? OFF : ON;
	led_cont(&led_green, tmp);
	mod_timer(led_green.timer, jiffies + HZ/2);
}

#define led_red_on()            \
        (red_enable   ? led_on(&led_red)            : 0)
#define led_red_off()           \
        (red_enable   ? led_off(&led_red)           : 0)
#define led_red_status()        \
        (red_enable   ? led_status(&led_red)        : 0)
#define led_red_blinkon()       \
        (red_enable   ? led_blinkon(&led_red)       : 0)
#define led_red_blinkoff()      \
        (red_enable   ? led_blinkoff(&led_red)      : 0)
#define led_red_blinkstatus()      \
        (red_enable   ? led_blinkstatus(&led_red)   : 0)

#define led_green_on()          \
        (green_enable ? led_on(&led_green)          : 0)
#define led_green_off()         \
        (green_enable ? led_off(&led_green)         : 0)
#define led_green_status()      \
        (green_enable ? led_status(&led_green)      : 0)
#define led_green_blinkon()     \
        (green_enable ? led_blinkon(&led_green)     : 0)
#define led_green_blinkoff()    \
        (green_enable ? led_blinkoff(&led_green)    : 0)
#define led_green_blinkstatus() \
        (green_enable ? led_blinkstatus(&led_green) : 0)

static int led_open(struct inode *inode, struct file *file)
{
	KDEBUG("led_open\n");

	spin_lock_irq(&led_lock);

	if(led_rivalstatus & LED_IS_OPEN){
		spin_unlock_irq(&led_lock);
		return -EBUSY;
	}

	led_rivalstatus = LED_IS_OPEN;
	spin_unlock_irq(&led_lock);
	return 0;
}

static int led_release(struct inode *inode, struct file *file)
{
	KDEBUG("led_release\n");

	spin_lock_irq(&led_lock);
	led_rivalstatus = 0;
	spin_unlock_irq(&led_lock);
	
	return 0;
}

static ssize_t led_read(struct file *file, char __user *buf,
                        size_t count, loff_t *ppos)
{
	unsigned long tmp;
	KDEBUG("led_read\n");

	if(count < 1) return -EINVAL;

	spin_lock_irq(&led_lock);
	if(led_rivalstatus & LED_READ_COMP){
		spin_unlock_irq(&led_lock);
		return 0;
	}
	
	led_rivalstatus |= LED_READ_COMP;
	spin_unlock_irq(&led_lock);
	
	tmp =  (led_red_status() ? LED_RED : 0);
	tmp |= (led_green_status() ? LED_GREEN : 0);
	
	tmp = sprintf(buf, "%1lx", tmp);
	return tmp;
}

static ssize_t led_write(struct file *file, const char __user *buf,
			 size_t count, loff_t *ppos)
{
	unsigned long tmp;
	KDEBUG("led_write\n");

	if(count < 1) return -EINVAL;

	spin_lock_irq(&led_lock);
	if(led_rivalstatus & LED_WRITE_COMP){
		spin_unlock_irq(&led_lock);
		return -EFBIG;
	}

	led_rivalstatus |= LED_WRITE_COMP;
	spin_unlock_irq(&led_lock);
	
	tmp = simple_strtol(buf,NULL,16);

	if(tmp & LED_RED) led_red_on();
	else led_red_off();

	if(tmp & LED_GREEN) led_green_on();
	else led_green_off();

	return count;
}

#define check_cmd(CMD,VALUE) ((CMD & VALUE) == VALUE ? 1 : 0 )

static int led_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
                     unsigned long arg)
{
	int flag = 0;
	int init_flag = 0;
	char *buf = (char *)arg;
	KDEBUG("led_ioctl\n");

	if(check_cmd(cmd,LED_RED_BLINK)) {
		if(check_cmd(cmd,LED_RED_ON)) { led_red_blinkon(); flag++; }
		if(check_cmd(cmd,LED_RED_OFF)) { led_red_blinkoff(); flag++; }
		if(check_cmd(cmd,LED_RED_STATUS)) {
			if(!init_flag) { buf[0] = 0; init_flag++; }
			buf[0] |= (led_red_blinkstatus() ? LED_RED : 0);
			flag++;
		}
	}
	else {
		if(check_cmd(cmd,LED_RED_ON)) { led_red_on(); flag++; }
		if(check_cmd(cmd,LED_RED_OFF)) { led_red_off(); flag++; }
		if(check_cmd(cmd,LED_RED_STATUS)) {
			if(!init_flag) { buf[0] = 0; init_flag++; }
			buf[0] |= (led_red_status() ? LED_RED : 0);
			flag++;
		}
	}
	if(check_cmd(cmd,LED_GREEN_BLINK)) {
		if(check_cmd(cmd,LED_GREEN_ON))
			{ led_green_blinkon(); flag++; }
		if(check_cmd(cmd,LED_GREEN_OFF))
			{ led_green_blinkoff(); flag++; }
		if(check_cmd(cmd,LED_GREEN_STATUS)) {
			if(!init_flag) { buf[0] = 0; init_flag++; }
			buf[0] |= (led_green_blinkstatus() ? LED_GREEN : 0);
			flag++;
		}
	}
	else {
		if(check_cmd(cmd,LED_GREEN_ON)) { led_green_on(); flag++; }
		if(check_cmd(cmd,LED_GREEN_OFF)) { led_green_off(); flag++; }
		if(check_cmd(cmd,LED_GREEN_STATUS)) {
			if(!init_flag) { buf[0] = 0; init_flag++; }
			buf[0] |= (led_green_status() ? LED_GREEN : 0);
			flag++;
		}
	}

	if(!flag) return -EINVAL;

	return 0;
}


static struct file_operations led_fops = {
	.owner   = THIS_MODULE,
	.open    = led_open,
	.release = led_release,
	.read    = led_read,
	.write   = led_write,
	.ioctl   = led_ioctl,
};

static struct miscdevice led_dev = {
	.minor = LED_MINOR,
	.name  = LED_DEVICE,
	.fops  = &led_fops,
};

static int __init led_init(void)
{
	KDEBUG("led_init\n");
	if(misc_register(&led_dev)){
		return -ENODEV;
	}

	printk(KERN_INFO VERSION "\n");

	check_platform();

	led_red.status = led_cont(&led_red, STATUS) ? ON : OFF;
	led_green.status = led_cont(&led_green, STATUS) ? ON : OFF;

	return 0;
}

static void __exit led_exit(void)
{
	KDEBUG("led_exit\n");
	misc_deregister(&led_dev);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(led_init);
module_exit(led_exit);
